"""------------------------------------------------------------------------------------------
Bayesian Affect Control Theory
Discrete Tutor Subclass of Agent class
Author: Jesse Hoey  jhoey@cs.uwaterloo.ca   http://www.cs.uwaterloo.ca/~jhoey
September 2013
Use for research purposes only.
Please do not re-distribute without written permission from the author
Any commerical uses strictly forbidden.
Code is provided without any guarantees.
Research sponsored by the Natural Sciences and Engineering Council of Canada (NSERC).
use python2.6
see README for details
----------------------------------------------------------------------------------------------"""

from bayesact import *


class DiscreteTutoringAgent(Agent):
    def __init__(self,*args,**kwargs):

        super(DiscreteTutoringAgent, self).__init__(self,*args,**kwargs)
        #x for the DiscreteTutoringAgent is a triple: 
        #the first value is the turn (always this way for any Agent)
        #the second value
        #is the difficulty level (0-2) of the previous exercise, and
        #the third is the skill level of the student
        #skill levels
        self.xvals=["easy","medium","hard"]
        #probability distribution over skill levels
        self.px=kwargs.get("initx",[0.2,0.6,0.2])
        #basic distributions for deflection of 1
        self.ppx=kwargs.get("ppx",[[0.9,0.1,0.0],[0.05,0.9,0.05],[0.0,0.1,0.9]])
        #prob. of failure/success for each difficulty level: easy, medium, hard
        #for each value of x[2]:   easy,med,hard
        self.of=kwargs.get("of",[[[0.1,0.9],[0.01,0.99],[0.001,0.999]],#x[1]=easy
                                 [[0.5,0.5],[0.1,0.9],[0.01,0.99]],#x[1]=med
                                 [[0.9,0.1],[0.5,0.5],[0.1,0.9]]])#x[1]=hard
        self.goalx=kwargs.get("goalx",2)

        self.x_avg=[0,0,0]


    def print_params(self):
        Agent.print_params(self)
        print "px: ",self.px
        print "ppx: ",self.ppx
        print "goal x: ",self.goalx

    #The only difference between a TutoringAgent and an Agent
    #is dynamics of x and reward
    def initialise_x(self,initx=None):
        if not initx:
            initpx=self.px
            initturn=State.turnnames.index("agent")
        else:
            initpx=initx[1]
            initturn=initx[0]
        initskill = list(NP.random.multinomial(1,initpx)).index(1)
        return [initturn,0,initskill]

    #implement a default policy where we look at the client's current abilities  (average x value)
    #and pick a difficulty level corresponding to P(x) - the distribution over x for client
    #select an action at the difficulty level corresponding to the client's ability
    #and occasionally one that is one level higher
    #avgstate=sampleMean(self.samples)  #will be a real number from 0-2
    def get_prop_action(self,skill=0):
	skill = self.x_avg[2]
        propositional_action=int(round(skill))
        if NP.random.random()>0.9:
            propositional_action=min(2,propositional_action+1)
        return propositional_action

    #was greedy_action_select
    #selects an action greedily by 
    def get_next_action(self,state,paab):
	N=50
	M=20
        fvsamples=[]
        max_rrew=0
        best_action=[]
        for x in range(N):
            (tmpfv,wgt,tmpH,tmpC)=sampleFvar(fvars,tvars,self.iH,self.iC,self.isiga,self.isigf_unconstrained_b,state.tau,state.f,state.get_turn(),[],[])
            tmpfv=tmpfv.transpose()
            aab=map(lambda x: float(x), tmpfv[3:6])
            #simulate next state given that action
            # here we should really average over all the self.samples, not over just the single average state passed in
            rrew=0.0
            for y in range(M):
                sample=self.sampleNext(fvars,tvars,state,state.get_turn(),aab,[],paab)
                #get reward of that -
                rrew += self.reward(sample)
            rrew = rrew/M
            if x==0 or rrew>max_rrew:
                max_rrew = rrew
                best_action=aab
        print "best rew: ",max_rrew
        print "best action: ",best_action
        return best_action



    #aab is the affective part of the action, paab is the propositional part
    def sampleXvar(self,f,tau,state,aab,paab):
        if state.get_turn()=="client":
            #compute deflection between f and tau
            #x moves towards goalx inversely proportionally to 
            #the deflection
            D=NP.dot(f-tau,f-tau)
            #as deflection grows, the distribtuion is more likely to favor staying the same or regressing, not progressing
            xppx=NP.asarray(self.ppx[state.x[2]])
            #current and all previous ones as wellget increased by deflection
            for y in range(state.x[2]+1):
                xppx[y] = xppx[y]*(D/4.0)
            #y=x[1]
            xppx = xppx/float(xppx.sum())
            #sample from this to get the new x value
            new_x=list(NP.random.multinomial(1,xppx)).index(1)
            #print xppx
            #print new_x
            return [state.invert_turn(),state.x[1],new_x]
        else:
            return [state.invert_turn(),paab,state.x[2]]
    
    #this must be in the class
    def evalSampleXvar(self,sample,xobs):
        #this nasty little hack is because I'm calling update(..) to update the agent
        #in tutorgui2.py, which makes a default assumption that on agent turn, no observations
        #are received. 
        #if xobs==[] or not turn=="client":
        #    return 1.0
        #print xobs[0],sample.x[0]
        if xobs[0]==sample.x[0]:
            #print sample.x[0],sample.x[1],xobs,self.of[sample.x[0]][sample.x[1]][xobs]
            return self.of[sample.x[1]][sample.x[2]][xobs[1]]
        else:
            return 0.0

    def reward(self,sample):
        # a generic deflection-based reward that can be removed eventually
        fsample=sample.f.transpose()
        freward = -1.0*NP.dot(sample.f-sample.tau,sample.f-sample.tau)
        # a state-based reward that favors states closer to the goal
        xreward = -1.0*(sample.x[2]-self.goalx)**2

        return freward+xreward
